/*
 * QEMU model of the EFuse_Cache
 *
 * Copyright (c) 2017 Xilinx Inc.
 *
 * Partially generated by xregqemu.py 2017-06-05.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"
#include "hw/sysbus.h"
#include "qemu/bitops.h"
#include "qemu/log.h"
#include "hw/block/xlnx-efuse.h"

#ifndef XILINX_EFUSE_CACHE_ERR_DEBUG
#define XILINX_EFUSE_CACHE_ERR_DEBUG 0
#endif

#define DPRINT(...) \
    if (XILINX_EFUSE_CACHE_ERR_DEBUG) {  \
        qemu_log(__VA_ARGS__);         \
    }

#define TYPE_XILINX_EFUSE_CACHE "xlnx.pmc-efuse-cache"

#define XILINX_EFUSE_CACHE(obj) \
     OBJECT_CHECK(EFuseCache, (obj), TYPE_XILINX_EFUSE_CACHE)

#define DPRINT_GE(args, ...) \
     qemu_log_mask(LOG_GUEST_ERROR, "%s: " args, __func__, ## __VA_ARGS__);

#define R_HOLE_START 0x30
#define R_HOLE_END   0x8C
#define V_HOLE       0

#define R_MAX 0xFFC

typedef struct EFuseCache {
    SysBusDevice parent_obj;
    MemoryRegion iomem;

    XLNXEFuse *efuse;
} EFuseCache;

static uint64_t efuse_cache_read(void *opaque, hwaddr addr, unsigned size)
{
    EFuseCache *s = XILINX_EFUSE_CACHE(opaque);
    unsigned int start = addr * 8;
    unsigned int end = (addr + size) * 8;
    uint32_t *ret = g_malloc0(size * 8 * sizeof(uint32_t));

    DPRINT("%s: addr=>0x%x ", __func__, (unsigned int)addr);
    switch (addr) {
    case R_HOLE_START ... R_HOLE_END:
        DPRINT_GE("Should not access unreadable efuse-cache hole %#llx!\n",
                  (unsigned long long)addr);
        return V_HOLE;
    default:
        break;
    };
    efuse_sync_u32(s->efuse, ret, start, end, FBIT_UNKNOWN);

    DPRINT("val=>0x%x\n", ret[0]);
    return ret[0];
}

static void efuse_cache_write(void *opaque, hwaddr addr, uint64_t value,
                              unsigned size)
{
    /* No Register Writes allowed */
    DPRINT_GE("Invalid write to efuse cache registers");
}

static const MemoryRegionOps efuse_cache_ops = {
    .read = efuse_cache_read,
    .write = efuse_cache_write,
    .endianness = DEVICE_LITTLE_ENDIAN,
    .valid = {
        .min_access_size = 4,
        .max_access_size = 4,
    },
};

static void efuse_cache_init(Object *obj)
{
    EFuseCache *s = XILINX_EFUSE_CACHE(obj);
    SysBusDevice *sbd = SYS_BUS_DEVICE(obj);

    memory_region_init_io(&s->iomem, obj, &efuse_cache_ops, s,
                          TYPE_XILINX_EFUSE_CACHE, R_MAX);
    sysbus_init_mmio(sbd, &s->iomem);
    object_property_add_link(obj, "efuse", TYPE_XLNX_EFUSE,
                             (Object **)&s->efuse,
                             qdev_prop_allow_set_link_before_realize,
                             OBJ_PROP_LINK_UNREF_ON_RELEASE,
                             NULL);
}

static const VMStateDescription vmstate_efuse_cache = {
    .name = TYPE_XILINX_EFUSE_CACHE,
    .version_id = 1,
    .minimum_version_id = 1,
    .fields = (VMStateField[]) {
        VMSTATE_END_OF_LIST(),
    }
};

static void efuse_cache_class_init(ObjectClass *klass, void *data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);

    dc->vmsd = &vmstate_efuse_cache;
}

static const TypeInfo efuse_cache_info = {
    .name          = TYPE_XILINX_EFUSE_CACHE,
    .parent        = TYPE_SYS_BUS_DEVICE,
    .instance_size = sizeof(EFuseCache),
    .class_init    = efuse_cache_class_init,
    .instance_init = efuse_cache_init,
};

static void efuse_cache_register_types(void)
{
    type_register_static(&efuse_cache_info);
}

type_init(efuse_cache_register_types)
